# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for ContextProvider classes.

Input can be a single observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import re
from abc import abstractmethod
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.utility import export
from ..lookup_result import SanitizedObservable
from ..provider_base import Provider

__version__ = VERSION
__author__ = "Ian Hellen"

# Regular expression from Grok patterns
# https://github.com/hpcugent/logstash-patterns/blob/master/files/grok-patterns
HOSTNAME_REGEX = (
    r"\b(?:[0-9A-Za-z][0-9A-Za-z-]{0,62})"
    r"(?:\.(?:[0-9A-Za-z][0-9A-Za-z-]{0,62}))*(\.?|\b)"
)


def _validate_hostname(hostname: str, **kwargs) -> SanitizedObservable:
    """Validate that parameter is a valid hostname."""
    del kwargs
    match_hostname = re.compile(HOSTNAME_REGEX, re.I | re.X | re.M).search(hostname)
    if not match_hostname:
        return SanitizedObservable(None, "Unrecognized hostname")

    return SanitizedObservable(match_hostname.group(0), "ok")


def _validate_ip(ipaddress: str, **kwargs):
    """Ensure Ip address is a valid public IPv4 address."""
    version = kwargs.pop("version", 4)
    try:
        addr = ip_address(ipaddress)
    except ValueError:
        return SanitizedObservable(None, "IP address is invalid format")

    if version == 4 and not isinstance(addr, IPv4Address):
        return SanitizedObservable(None, "Not an IPv4 address")
    if version == 6 and not isinstance(addr, IPv6Address):
        return SanitizedObservable(None, "Not an IPv6 address")

    return SanitizedObservable(ipaddress, "ok")


@export
class ContextProvider(Provider):
    """Abstract base class for Context providers."""

    _REQUIRED_PARAMS: List[str] = []

    def __init__(self, **kwargs) -> None:
        """Initialize Context provider."""
        super().__init__(**kwargs)
        self._preprocessors._processors.pop("hostname")
        self._preprocessors._processors.pop("ipv4")
        self._preprocessors._processors.pop("ipv6")
        self._preprocessors.add_check("hostname", _validate_hostname)
        self._preprocessors.add_check("ipv4", _validate_ip)
        self._preprocessors.add_check("ipv6", _validate_ip)

    def lookup_item(
        self,
        item: str,
        item_type: Optional[str] = None,
        query_type: Optional[str] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup from a value.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            The Type of the item to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular observable to try avoid repeated network calls for
        the same item.

        """
        return self.lookup_observable(
            observable=item,
            observable_type=item_type,
            query_type=query_type,
            **kwargs,
        )

    @abstractmethod
    def lookup_observable(  # type: ignore
        self,
        observable: str,
        observable_type: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a single observable.

        Parameters
        ----------
        observable : str
            Observable value to lookup
        observable_type : str, optional
            The Type of the value to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The context lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        """

    def _check_item_type(
        self, item: str, item_type: str = None, query_subtype: str = None
    ) -> Dict:
        """
        Check Item Type and cleans up item.

        Parameters
        ----------
        item : str
            item
        item_type : str, optional
            item type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed item.
            Status is none-zero on failure.

        """
        return self._check_observable_type(
            item,
            item_type,
            query_subtype,
        )

    def _check_observable_type(
        self, obs: str, obs_type: str = None, query_subtype: str = None
    ) -> Dict:
        """
        Check Observable Type and cleans up observable.

        Parameters
        ----------
        obs : str
            Observable
        obs_type : str, optional
            Observable type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed Observable.
            Status is none-zero on failure.

        """
        result = super()._check_item_type(
            item=obs, item_type=obs_type, query_subtype=query_subtype
        )
        result["Observable"] = result.pop("Item")
        result["ObservableType"] = result.pop("ItemType")
        result["SafeObservable"] = result.pop("SanitizedValue")
        return result

    @abstractmethod
    def parse_results(self, response: Dict) -> Tuple[bool, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """

    def lookup_observables(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup collection of observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observables
            3. Iterable of observables
        obs_col : str, optional
            DataFrame column to use for observables, by default None
        obs_type_col : str, optional
            DataFrame column to use for observables types, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return self.lookup_items(
            data,
            item_col=obs_col,
            item_type_col=obs_type_col,
            query_type=query_type,
            **kwargs,
        )

    async def lookup_observables_async(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Call base async wrapper."""
        return await self._lookup_items_async_wrapper(
            data,
            item_col=obs_col,
            item_type_col=obs_type_col,
            query_type=query_type,
            **kwargs,
        )

    async def _lookup_observables_async_wrapper(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Async wrapper for providers that do not implement lookup_iocs_async.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, Observables Types
            3. Iterable of observables - Observables Types will be inferred
        obs_col : str, optional
            DataFrame column to use for observables, by default None
        obs_type_col : str, optional
            DataFrame column to use for Observables Types, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return await self._lookup_items_async_wrapper(
            data,
            item_col=obs_col,
            item_type_col=obs_type_col,
            query_type=query_type,
            **kwargs,
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Checker functions for Azure ML notebooks."""
import logging
import os
import sys
import warnings
from pathlib import Path
from typing import Any, List, Mapping, Optional, Tuple, Union

import yaml
from IPython import get_ipython
from IPython.display import HTML, display
from pkg_resources import Requirement, WorkingSet, parse_version  # type: ignore

from .._version import VERSION
from ..common.pkg_config import _HOME_PATH, get_config, refresh_config
from ..common.utility import search_for_file, unit_testing
from ..config import MpConfigFile  # pylint: disable=no-name-in-module

__version__ = VERSION

logger = logging.getLogger(__name__)

AZ_GET_STARTED = (
    "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/A%20Getting"
    "%20Started%20Guide%20For%20Azure%20Sentinel%20ML%20Notebooks.ipynb"
)
TROUBLE_SHOOTING = (
    "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/"
    "TroubleShootingNotebooks.ipynb"
)
MISSING_PKG_ERR = """
    <h4><font color='orange'>The package '<b>{package}</b>' is not
    installed or has an unsupported version (installed version = '{inst_ver}')</font></h4>
    Please install or upgrade before continuing: required version is {package}>={req_ver}
    """
MP_INSTALL_FAILED = """
    <h4><font color='red'>The notebook may not run correctly without
    the correct version of '<b>{pkg}</b>' ({ver} or later).</font></h4>
    Please see the <a href="{nbk_uri}">
    Getting Started Guide For Azure Sentinel ML Notebooks</a></b>
    for more information<br><hr>
"""
RELOAD_MP = """
    <h4><font color='orange'>Kernel restart needed</h4>
    An error was detected trying to load the updated version of MSTICPy.<br>
    Please restart the notebook kernel and re-run this cell - it should
    run without error.
    """
_AZ_CLI_WIKI_URI = (
    "https://github.com/Azure/Azure-Sentinel-Notebooks/wiki/"
    "Caching-credentials-with-Azure-CLI"
)
_CLI_WIKI_MSSG_GEN = (
    f"For more information see <a href='{_AZ_CLI_WIKI_URI}'>"
    "Caching credentials with Azure CLI</>"
)
_CLI_WIKI_MSSG_SHORT = (
    f"see <a href='{_AZ_CLI_WIKI_URI}'>Caching credentials with Azure CLI</>"
)
_PYSPARK_KERNEL_NOT_SUPPORTED = """
    <h4><font color='red'>WARNING: This notebook may not run correctly with the Azure Machine
    Learning PySpark kernel.</font></h4> Please create a standard Azure Machine Learning
    Compute instance and select the Python 3.10 or later kernel.
    Please see the <a href="{nbk_uri}">
    Getting Started Guide For Azure Sentinel ML Notebooks</a></b>
    for more information<br><hr>
"""

MIN_PYTHON_VER_DEF = "3.10.0"
MSTICPY_REQ_VERSION = __version__

VER_RGX = r"(?P<maj>\d+)\.(?P<min>\d+).(?P<pnt>\d+)(?P<suff>.*)"
MP_ENV_VAR = "MSTICPYCONFIG"
MP_FILE = "msticpyconfig.yaml"


def is_in_aml_pyspark():
    """Return True if running in Spark on Azure Machine Learning."""
    return os.environ.get("AZUREML_FRAMEWORK") == "PySpark"


def is_in_aml():
    """Return True if running in Azure Machine Learning."""
    return os.environ.get("APPSETTING_WEBSITE_SITE_NAME") == "AMLComputeInstance"


def check_aml_settings(
    min_py_ver: Union[str, Tuple] = MIN_PYTHON_VER_DEF,
    min_mp_ver: Union[str, Tuple] = MSTICPY_REQ_VERSION,
    extras: Optional[List[str]] = None,
    mp_release: Optional[str] = None,
    **kwargs,
):
    """
    Check the current versions of the Python kernel and MSTICPy.

    Parameters
    ----------
    min_py_ver : Union[Tuple[int, int], str]
        Minimum Python version
    min_mp_ver : Union[Tuple[int, int], str]
        Minimum MSTICPy version
    extras : Optional[List[str]], optional
        A list of extras required for MSTICPy
    mp_release : Optional[str], optional
        Override the MSTICPy release version. This
        can also be specified in the environment variable 'MP_TEST_VER'

    Raises
    ------
    RuntimeError
        If the Python version does not support the notebook.
        If the MSTICPy version does not support the notebook
        and the user chose not to upgrade

    """
    del kwargs
    _disp_html("<h4>Starting AML notebook pre-checks...</h4>")
    _check_pyspark()
    if isinstance(min_py_ver, str):
        min_py_ver = _get_pkg_version(min_py_ver).release  # type: ignore
    check_python_ver(min_py_ver=min_py_ver)

    _check_mp_install(min_mp_ver, mp_release, extras)
    if _kql_magic_installed():
        _check_kql_prereqs()
        _set_kql_env_vars(extras)
    _run_user_settings()
    _set_mpconfig_var()
    _check_azure_cli_status()
    _check_aml_auth_method_order()
    _disp_html("<h4>Notebook pre-checks complete.</h4>")


# retain previous name for backward compatibility
check_versions = check_aml_settings


def _check_pyspark():
    """Check if we are running in PySpark kernel."""
    if not is_in_aml_pyspark():
        return

    _disp_html(_PYSPARK_KERNEL_NOT_SUPPORTED.format(nb_uri=AZ_GET_STARTED))


def _kql_magic_installed():
    try:
        # pylint: disable=import-outside-toplevel, unused-import
        from Kqlmagic import kql  # noqa: F401

        return True
    except ImportError:
        return False


def check_python_ver(min_py_ver: Union[str, Tuple] = MIN_PYTHON_VER_DEF):
    """
    Check the current version of the Python kernel.

    Parameters
    ----------
    min_py_ver : Tuple[int, int]
        Minimum Python version

    Raises
    ------
    RuntimeError
        If the Python version does not support the notebook.

    """
    minimum_py_version = max(
        _get_pkg_version(min_py_ver), _get_pkg_version(MIN_PYTHON_VER_DEF)
    )
    sys_ver = _get_pkg_version(sys.version_info[:3])
    _disp_html("Checking Python kernel version...")
    if sys_ver < minimum_py_version:
        # Bandit SQL inject error found here
        _disp_html(
            f"""
            <h4><font color='red'>This notebook requires a later
            (Python) kernel version.</h4></font>
            Select a kernel from the notebook toolbar (above), that is Python
            {minimum_py_version} or later (Python 3.10 recommended)<br>
            """
        )  # nosec
        _disp_html(
            f"""
            Please see the <a href="{TROUBLE_SHOOTING}">TroubleShootingNotebooks</a>
            for more information<br><br><hr>
            """
        )
        # Bandit SQL inject error found here
        raise RuntimeError(
            f"Python {minimum_py_version} or later kernel is required."
        )  # nosec

    _disp_html(f"Info: Python kernel version {sys_ver} - OK<br>")


def check_mp_ver(min_msticpy_ver: Union[str, Tuple], extras: Optional[List[str]]):
    """
    Check and optionally update the current version of msticpy.

    Parameters
    ----------
    min_msticpy_ver : Tuple[int, int]
        Minimum MSTICPy version
    extras : Optional[List[str]], optional
        A list of extras required for MSTICPy

    Raises
    ------
    ImportError
        If MSTICPy version is insufficient and we need to upgrade

    """
    mp_min_pkg_ver = _get_pkg_version(min_msticpy_ver)

    _disp_html("Checking msticpy version...<br>")

    # Check currently loaded MP version against notebook requirement
    loaded_version = _get_pkg_version(__version__)
    if loaded_version < mp_min_pkg_ver:
        _disp_html(
            MISSING_PKG_ERR.format(
                package="msticpy",
                inst_ver=loaded_version,
                req_ver=mp_min_pkg_ver,
            )
        )
        mp_pkg_spec = f"msticpy[{','.join(extras)}]" if extras else "msticpy"
        mp_pkg_spec = f"{mp_pkg_spec}>={min_msticpy_ver}"

        _disp_html(
            f"Please run the following command to upgrade MSTICPy<br>"
            f"<pre>%pip install --upgrade {mp_pkg_spec}</pre><br>"
        )
        raise ImportError(
            "Unsupported version of MSTICPy installed",
            f"Installed version: {loaded_version}",
            f"Required version: {mp_min_pkg_ver}",
        )

    # Check loaded version against installed version in the environment
    # If the version was updated after this version was loaded by Python
    # we need to warn the user to restart the kernel.
    installed_version = _get_installed_mp_version()
    if installed_version and installed_version > loaded_version:
        _disp_html(
            f"A newer version of MSTICPy ({installed_version})"
            "has been installed but has not been loaded.<br>"
            "Please restart the notebook kernel and re-run this cell."
        )
    _disp_html(f"Info: msticpy version {loaded_version} (>= {mp_min_pkg_ver}) - OK<br>")


def populate_config_to_mp_config(mp_path):
    """Populate new or existing msticpyconfig with settings from config.json."""
    # Look for a config.json
    config_json = search_for_file("config.json", paths=[get_aml_user_folder()])
    if not config_json:
        return None

    # if we found one, use it to populate msticpyconfig.yaml
    mp_path = mp_path or str(get_aml_user_folder().joinpath("msticpyconfig.yaml"))
    mp_config_convert = MpConfigFile(file=config_json)
    azs_settings = mp_config_convert.map_json_to_mp_ws()
    def_azs_settings = next(
        iter(azs_settings.get("AzureSentinel", {}).get("Workspaces", {}).values())
    )
    if def_azs_settings:
        mp_config_convert.settings["AzureSentinel"]["Workspaces"][
            "Default"
        ] = def_azs_settings.copy()

    if Path(mp_path).exists():
        # If there is an existing file read it in
        mp_config_text = Path(mp_path).read_text(encoding="utf-8")
        mp_config_settings = yaml.safe_load(mp_config_text)
        # update exist settings with the AzSent settings from config.json
        mp_config_settings.update(mp_config_convert.settings)
        # update MpConfigFile with the merged settings
        mp_config_convert.settings = mp_config_settings
        mssg = (
            f"Updated '{mp_path}' with Microsoft Sentinel settings"
            f" imported from {config_json}."
        )
    else:
        mssg = (
            f"Created '{mp_path}' with Microsoft Sentinel settings"
            f" imported from {config_json}."
        )
    # Save the file
    mp_config_convert.save_to_file(mp_path, backup=True)
    logger.info(mssg)
    return mssg


def _check_mp_install(
    min_mp_ver: Union[str, Tuple],
    mp_release: Optional[str],
    extras: Optional[List[str]],
):
    """Check for and try to install required MSTICPy version."""
    # Use the release ver specified in params, in the environment or
    # the notebook default.
    pkg_version = _get_pkg_version(min_mp_ver)
    mp_install_version = mp_release or os.environ.get("MP_TEST_VER") or str(pkg_version)

    check_mp_ver(min_msticpy_ver=mp_install_version, extras=extras)


def _set_kql_env_vars(extras: Optional[List[str]]):
    """Set environment variables for Kqlmagic based on MP extras."""
    jp_extended = ("azsentinel", "azuresentinel", "kql")
    if extras and any(extra for extra in extras if extra in jp_extended):
        os.environ["KQLMAGIC_EXTRAS_REQUIRE"] = "jupyter-extended"
    else:
        os.environ["KQLMAGIC_EXTRAS_REQUIRE"] = "jupyter-basic"
    if is_in_aml():
        os.environ["KQLMAGIC_AZUREML_COMPUTE"] = _get_vm_fqdn()


def _get_pkg_version(version: Union[str, Tuple]):
    """Return pkg_resources parsed version from string or tuple."""
    if isinstance(version, str):
        return parse_version(version)
    if isinstance(version, tuple):
        return parse_version(".".join(str(ver) for ver in version))
    raise TypeError(f"Version {version} not cannot be parsed.")


def _get_installed_mp_version():
    """Return the installed version of MSTICPY."""
    working_set = WorkingSet()
    mp_installed = working_set.find(Requirement("msticpy"))  # type: ignore
    if mp_installed:
        return mp_installed.parsed_version
    return None


def _disp_html(text: str):
    """Display the HTML text."""
    display(HTML(text))


def get_aml_user_folder() -> Optional[Path]:
    """Return the root of the user folder."""
    path_parts = Path(".").absolute().parts
    if "Users" not in path_parts:
        return Path(_HOME_PATH).expanduser()
    # find the index of the last occurrence of "users"
    users_idx = len(path_parts) - path_parts[::-1].index("Users")
    # the user folder is one item below this
    if len(path_parts) < users_idx + 1:
        return None
    return Path("/".join(path_parts[: users_idx + 1]))


# pylint: disable=import-outside-toplevel, unused-import, import-error
def _run_user_settings():
    """Check for nbuser_settings.py and warn that it is no longer used."""
    user_folder = get_aml_user_folder()
    if user_folder and user_folder.joinpath("nbuser_settings.py").is_file():
        warnings.warn(
            "Automatic processing of nbuser_settings.py is no longer supported in MSTICPy."
        )


# pylint: enable=import-outside-toplevel, unused-import, import-error


def _set_mpconfig_var():
    """Set MSTICPYCONFIG to file in user directory if no other found."""
    mp_path_val = os.environ.get(MP_ENV_VAR)
    if (
        # If a valid MSTICPYCONFIG value is found - return
        (mp_path_val and Path(mp_path_val).is_file())
        # Or if there is a msticpyconfig in the current folder.
        or Path(".").joinpath(MP_FILE).is_file()
    ):
        return
    # Otherwise check the user's root folder
    user_dir = get_aml_user_folder()
    if user_dir:
        mp_path = Path(user_dir).joinpath(MP_FILE)
        if mp_path.is_file():
            # If there's a file there, set the env variable to that.
            os.environ[MP_ENV_VAR] = str(mp_path)
            # Since we have already imported msticpy to check the version
            # it will have already configured settings so we need to refresh.
            refresh_config()
            _disp_html(
                f"<br>No {MP_FILE} found. Will use {MP_FILE} in user folder {user_dir}<br>"
            )


_NBVM_PATH = "/mnt/azmnt/.nbvm"


def _get_vm_metadata() -> Mapping[str, Any]:
    """Read VM metadata from definition file."""
    with open(_NBVM_PATH, "r", encoding="utf-8") as nbvm_handle:
        nbvm_lines = nbvm_handle.readlines()
    return {
        item[0]: item[1]
        for item in map(lambda x: x.split("=", maxsplit=1), nbvm_lines)
        if item
    }


def _get_vm_fqdn() -> str:
    """Get the FQDN of the host."""
    vm_metadata = _get_vm_metadata()
    if vm_metadata and "instance" in vm_metadata:
        return (
            f"https://{vm_metadata.get('instance')}.{vm_metadata.get('domainsuffix')}"
        )
    return ""


def _check_kql_prereqs():
    """
    Check and install packages for Kqlmagic/msal_extensions.

    Notes
    -----
    Kqlmagic may trigger warnings about a missing PyGObject package
    and some system library dependencies. To fix this do the
    following:<br>
    From a notebook run:

        %pip uninstall enum34
        !sudo apt-get --yes install libgirepository1.0-dev
        !sudo apt-get --yes install gir1.2-secret-1
        %pip install pygobject

    You can also do this from a terminal - but ensure that you've
    activated the environment corresponding to the kernel you are
    using prior to running the pip commands.

        # Install the libgi dependency
        sudo apt install libgirepository1.0-dev
        sudo apt install gir1.2-secret-1

        # activate the environment
        # conda activate azureml_py38
        # source ./env_path/scripts/activate

        # Uninstall enum34
        python -m pip uninstall enum34
        # Install pygobject
        python -m install pygobject

    """
    if not is_in_aml():
        return
    try:
        # If this successfully imports, we are ok
        # pylint: disable=import-outside-toplevel
        import gi

        # pylint: enable=import-outside-toplevel
        del gi
    except ImportError:
        # Check for system packages
        ip_shell = get_ipython()
        if not ip_shell:
            return
        apt_list = ip_shell.run_line_magic("sx", "apt list")
        apt_list = [apt.split("/", maxsplit=1)[0] for apt in apt_list]
        missing_lx_pkg = [
            apt_pkg
            for apt_pkg in ("libgirepository1.0-dev", "gir1.2-secret-1")
            if apt_pkg not in apt_list
        ]
        if missing_lx_pkg:
            _disp_html(
                "Kqlmagic/msal-extensions pre-requisite PyGObject not installed."
            )
            _disp_html(
                "To prevent warnings when loading the Kqlmagic data provider,"
                " Please run the following command:<br>"
                "!conda install --yes -c conda-forge pygobject<br>"
            )


def _check_azure_cli_status():
    """Check for Azure CLI credentials."""
    # import these only if we need them at runtime
    # pylint: disable=import-outside-toplevel
    from ..auth.azure_auth_core import AzureCliStatus, check_cli_credentials

    if unit_testing():
        return

    status, message = check_cli_credentials()
    if status == AzureCliStatus.CLI_OK:
        _disp_html(message)
        return

    _disp_html(
        "<h4>Azure CLI authentication recommended</h4>"
        "Azure CLI gives the most seamless authentication experience for "
        "authenticating to Azure Resources from Azure Machine Learning.<br>"
        "Create a new notebook cell and enter the following command:<br>"
        "<pre>!az login</pre>"
        "You will be prompted to authenticate to Azure. Once you have "
        "completed this, return to this notebook and continue.<br>"
    )
    if status == AzureCliStatus.CLI_NOT_INSTALLED:
        _disp_html("Azure CLI not installed." f" ({_CLI_WIKI_MSSG_SHORT})")
    elif message:
        _disp_html("\n".join([message, _CLI_WIKI_MSSG_GEN]))


_MSI_WARNING = """
<h4><font color='red'>WARNING: MSI authentication to MS Sentinel not supported</font></h4>
Your have MSI authentication as higher priority than CLI or DeviceCode.
This may cause authentication to MS Sentinel to fail since MSI authentication from
Azure Machine Learning is no longer supported.<br>
We recommend that you change your msticpyconfig.yaml so that
you either remove MSI or place it below CLI and device code methods.<br>
Example:<br>
<pre>
Azure:
    auth_methods:
        - cli
        - devicecode
        - msi
</pre>
"""


def _check_aml_auth_method_order():
    """Reorder the auth methods to put Azure CLI first."""
    try:
        current_methods = get_config("Azure.auth_methods")
    except KeyError:
        return
    if not current_methods:
        return
    auth_index = {meth: idx for idx, meth in enumerate(current_methods)}
    msi_lower_than_cli = auth_index.get("msi", 99) > auth_index.get("cli", 99)
    msi_lower_than_devcode = auth_index.get("msi", 99) > auth_index.get(
        "devicecode", auth_index.get("device_code", 99)
    )
    if msi_lower_than_cli or msi_lower_than_devcode:
        return
    _disp_html(_MSI_WARNING)
    logging.warning("MSI authentication is higher priority than CLI or DeviceCode.")
    if "msi" in current_methods:
        _disp_html("Reordering auth_methods to move MSI to lowest priority.")
        current_methods.remove("msi")
        current_methods.append("msi")
        logging.info("Reordering auth_methods to move MSI to the end.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for ContextLookup classes.

Input can be a single Observable observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from typing import Dict, Iterable, List, Mapping, Optional, Union

import pandas as pd

from .._version import VERSION
from ..common.utility import export

# used in dynamic instantiation of providers
from .contextproviders import CONTEXT_PROVIDERS
from .lookup import Lookup
from .provider_base import Provider, _make_sync

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class ContextLookup(Lookup):
    """Observable lookup from providers."""

    _NO_PROVIDERS_MSG = """
    No Context Providers are loaded - please check that
    you have correctly configured your msticpyconfig.yaml settings.
    """

    _HELP_URI = (
        "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
        "ContextProviders.html#configuration-file"
    )

    PACKAGE = "contextproviders"

    PROVIDERS = CONTEXT_PROVIDERS
    CUSTOM_PROVIDERS: Dict[str, Provider] = {}

    # pylint: disable=too-many-arguments
    def lookup_observable(
        self,
        observable: str,
        observable_type: Optional[str] = None,
        query_type: Optional[str] = None,
        providers: Optional[List[str]] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup single observable in active providers.

        Parameters
        ----------
        observable : str
            observable
        observable_type : str, optional
            One of IoCExtract.IOCType, by default None
            If none, the Observable type will be inferred
        query_type: str, optional
            The observable query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]] = None,
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            The result returned as a DataFrame:
            bool indicates whether a TI record was found in any provider
            list has an entry for each provider result

        """
        return self.lookup_item(
            item=observable,
            item_type=observable_type,
            query_type=query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def lookup_observables(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of Observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Mapping (e.g. a dict) of [observable, ObservableType]
            3. Iterable of observables - ObservableTypes will be inferred
        obs_col : str, optional
            DataFrame column to use for observables, by default None
            ("col" and "column" are also aliases for this parameter)
        obs_type_col : str, optional
            DataFrame column to use for ObservableTypes, by default None
        query_type: str, optional
            The observable query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return _make_sync(
            self._lookup_observables_async(
                data=data,
                obs_col=obs_col,
                obs_type_col=obs_type_col,
                query_type=query_type,
                providers=providers,
                default_providers=default_providers,
                prov_scope=prov_scope,
                **kwargs,
            )
        )

    # pylint: disable=too-many-locals
    async def _lookup_observables_async(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """Lookup items async."""
        return await self._lookup_items_async(
            data,
            item_col=obs_col,
            item_type_col=obs_type_col,
            query_type=query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def lookup_observables_sync(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        obs_col: str = None,
        obs_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of IoCs.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Mapping (e.g. a dict) of [observable, IoCType]
            3. Iterable of observables - IoCTypes will be inferred
        obs_col : str, optional
            DataFrame column to use for observables, by default None
            ("col" and "column" are also aliases for this parameter)
        obs_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return self.lookup_items_sync(
            data,
            item_col=obs_col,
            item_type_col=obs_type_col,
            query_type=query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def _load_providers(self, **kwargs):
        """Load provider classes based on config."""
        return super()._load_providers(providers="ContextProviders", **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Geoip Lookup module using IPStack and Maxmind GeoLite2.

Geographic location lookup for IP addresses. This module has two classes
for different services:

-  GeoLiteLookup - Maxmind Geolite (see https://www.maxmind.com)
-  IPStackLookup - IPStack (see https://ipstack.com)

Both services offer
a free tier for non-commercial use. However, a paid tier will
normally get you more accuracy, more detail and a higher throughput
rate. Maxmind geolite uses a downloadable database, while IPStack is
an online lookup (API key required).

"""
import contextlib
import math
import random
import tarfile
import warnings
from abc import ABCMeta, abstractmethod
from collections import abc
from datetime import datetime, timedelta, timezone
from json import JSONDecodeError
from pathlib import Path
from time import sleep
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple

import geoip2.database  # type: ignore
import httpx
import pandas as pd
from geoip2.errors import AddressNotFoundError  # type: ignore
from IPython.display import HTML, display

from .._version import VERSION
from ..common.exceptions import MsticpyUserConfigError
from ..common.pkg_config import current_config_path, get_http_timeout
from ..common.provider_settings import ProviderSettings, get_provider_settings
from ..common.utility import SingletonArgsClass, export, is_ipython, mp_ua_header
from ..datamodel.entities import GeoLocation, IpAddress
from .ip_utils import get_ip_type

__version__ = VERSION
__author__ = "Ian Hellen"


class GeoIPDatabaseError(Exception):
    """Exception when GeoIP database cannot be found."""


class GeoIpLookup(metaclass=ABCMeta):
    """
    Abstract base class for GeoIP Lookup classes.

    See Also
    --------
    IPStackLookup : IPStack GeoIP Implementation
    GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _LICENSE_TXT: Optional[str] = None
    _LICENSE_HTML: Optional[str] = None

    @abstractmethod
    def lookup_ip(
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location abstract method.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]:
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """

    def df_lookup_ip(self, data: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Lookup Geolocation data from a pandas Dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            pandas dataframe containing IpAddress column
        column : str
            the name of the dataframe column to use as a source

        Returns
        -------
        pd.DataFrame
            Copy of original dataframe with IP Location information columns
            appended (where a location lookup was successful)

        """
        return data.merge(
            self.lookup_ips(data, column),
            how="left",
            left_on=column,
            right_on="IpAddress",
        )

    def lookup_ips(self, data: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Lookup Geolocation data from a pandas Dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            pandas dataframe containing IpAddress column
        column : str
            the name of the dataframe column to use as a source

        Returns
        -------
        pd.DataFrame
            IpLookup results as DataFrame.

        """
        ip_list = list(data[column].values)
        _, entities = self.lookup_ip(ip_addr_list=ip_list)

        ip_dicts = [
            {**ent.Location.properties, "IpAddress": ent.Address}
            for ent in entities
            if ent.Location is not None
        ]
        return pd.DataFrame(data=ip_dicts)

    @staticmethod
    def _ip_params_to_list(ip_address, ip_addr_list, ip_entity) -> List[str]:
        """Try to convert different parameter formats to list."""
        if ip_address is not None:
            # check if ip_address just used as positional arg.
            if isinstance(ip_address, str):
                return [ip_address.strip()]
            if isinstance(ip_address, abc.Iterable):
                return [str(ip).strip() for ip in ip_addr_list]
            if isinstance(ip_address, IpAddress):
                return [ip_entity.Address]
        if ip_addr_list is not None and isinstance(ip_addr_list, abc.Iterable):
            return [str(ip).strip() for ip in ip_addr_list]
        if ip_entity:
            return [ip_entity.Address]
        raise ValueError("No valid ip addresses were passed as arguments.")

    def print_license(self):
        """Print license information for providers."""
        if self._LICENSE_HTML and is_ipython(notebook=True):
            display(HTML(self._LICENSE_HTML))
        elif self._LICENSE_TXT:
            print(self._LICENSE_TXT)


@export
class IPStackLookup(GeoIpLookup):
    """
    IPStack GeoIP Implementation.

    See Also
    --------
    GeoIpLookup : Abstract base class
    GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _LICENSE_HTML = """
This library uses services provided by ipstack.
<a href="https://ipstack.com">https://ipstack.com</a>"""

    _LICENSE_TXT = """
This library uses services provided by ipstack (https://ipstack.com)"""

    _IPSTACK_API = "http://api.ipstack.com/{iplist}?access_key={access_key}&output=json"

    _NO_API_KEY_MSSG = """
No API Key was found to access the IPStack service.
If you do not have an account, go here to create one and obtain and API key.

Add this API key to your msticpyconfig.yaml
After adding the key run the following commands to reload your settings and retry:
    import msticpy
    msticpy.settings.refresh_config()

Alternatively, you can pass this to the IPStackLookup class when creating it:
>>> iplookup = IPStackLookup(api_key="your_api_key")
"""

    def __init__(self, api_key: Optional[str] = None, bulk_lookup: bool = False):
        """
        Create a new instance of IPStackLookup.

        Parameters
        ----------
        api_key : str, optional
            API Key from IPStack - see https://ipstack.com
            default is None - obtain key from msticpyconfig.yaml
        bulk_lookup : bool, optional
            For Professional and above tiers allowing you to
            submit multiple IPs in a single request.
            (the default is False, which submits a single request
            per address)

        """
        self.settings: Optional[ProviderSettings] = None
        self._api_key: Optional[str] = api_key
        self.bulk_lookup = bulk_lookup

    def _check_initialized(self):
        """Return True if valid API key available."""
        if self._api_key:
            return True

        self.settings = _get_geoip_provider_settings("IPStack")
        self._api_key = self.settings.args.get("AuthKey") if self.settings else None
        if self._api_key:
            return True
        raise MsticpyUserConfigError(
            self._NO_API_KEY_MSSG,
            help_uri=(
                "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
                + "GeoIPLookups.html#ipstack-geo-lookup-class"
            ),
            service_uri="https://ipstack.com/product",
            title="IPStack API key not found",
        )

    def lookup_ip(
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location from IPStack web service.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]:
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        Raises
        ------
        ConnectionError
            Invalid status returned from http request
        PermissionError
            Service refused request (e.g. requesting batch of addresses
            on free tier API key)

        """
        self._check_initialized()
        ip_list = self._ip_params_to_list(ip_address, ip_addr_list, ip_entity)

        results = self._submit_request(ip_list)
        output_raw = []
        output_entities = []
        for ip_loc, status in results:
            if status == 200 and "error" not in ip_loc:
                output_entities.append(self._create_ip_entity(ip_loc, ip_entity))
            output_raw.append((ip_loc, status))
        return output_raw, output_entities

    @staticmethod
    def _create_ip_entity(ip_loc: dict, ip_entity) -> IpAddress:
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_loc["ip"]
        geo_entity = GeoLocation()
        geo_entity.CountryCode = ip_loc["country_code"]

        geo_entity.CountryName = ip_loc["country_name"]
        geo_entity.State = ip_loc["region_name"]
        geo_entity.City = ip_loc["city"]
        geo_entity.Longitude = ip_loc["longitude"]
        geo_entity.Latitude = ip_loc["latitude"]
        if "connection" in ip_loc:
            geo_entity.Asn = ip_loc["connection"]["asn"]
        ip_entity.Location = geo_entity
        return ip_entity

    def _submit_request(self, ip_list: List[str]) -> List[Tuple[Dict[str, str], int]]:
        """
        Submit the request to IPStack.

        Parameters
        ----------
        ip_list : List[str]
            String list of IPs to look up

        Returns
        -------
        List[Tuple[str, int]]
            List of response, status code pairs

        """
        if not self.bulk_lookup:
            return self._lookup_ip_list(ip_list)

        submit_url = self._IPSTACK_API.format(
            iplist=",".join(ip_list), access_key=self._api_key
        )
        response = httpx.get(
            submit_url, timeout=get_http_timeout(), headers=mp_ua_header()
        )

        if response.status_code == 200:
            results = response.json()
            # {"success":false,"error":{"code":303,"type":"batch_not_supported_on_plan",
            # "info":"Bulk requests are not supported on your plan.
            # Please upgrade your subscription."}}

            if "success" in results and not results["success"]:
                raise PermissionError(
                    f"Service unable to complete request. Error: {results['error']}"
                )
            return [(item, response.status_code) for item in results]

        if response:
            with contextlib.suppress(JSONDecodeError):
                return [(response.json(), response.status_code)]
        return [({}, response.status_code)]

    def _lookup_ip_list(self, ip_list: List[str]):
        """Lookup IP Addresses one-by-one."""
        ip_loc_results = []
        with httpx.Client(timeout=get_http_timeout(), headers=mp_ua_header()) as client:
            for ip_addr in ip_list:
                submit_url = self._IPSTACK_API.format(
                    iplist=ip_addr, access_key=self._api_key
                )
                response = client.get(submit_url)
                if response.status_code == 200:
                    ip_loc_results.append((response.json(), response.status_code))
                elif response:
                    try:
                        ip_loc_results.append((response.json(), response.status_code))
                        continue
                    except JSONDecodeError:
                        ip_loc_results.append((None, response.status_code))
                else:
                    print("Unknown response from IPStack request.")
                    ip_loc_results.append((None, -1))
        return ip_loc_results


@export
@SingletonArgsClass
class GeoLiteLookup(GeoIpLookup):
    """
    GeoIP Lookup using MaxMindDB database.

    See Also
    --------
    GeoIpLookup : Abstract base class
    IPStackLookup : IPStack GeoIP Implementation

    """

    _MAXMIND_DOWNLOAD = (
        "https://download.maxmind.com/app/geoip_download?"
        + "edition_id=GeoLite2-City&license_key={license_key}&suffix=tar.gz"
    )

    _DB_HOME = str(Path.joinpath(Path("~").expanduser(), ".msticpy", "GeoLite2"))
    _DB_ARCHIVE = "GeoLite2-City.mmdb.{rand}.tar.gz"
    _DB_FILE = "GeoLite2-City.mmdb"

    _LICENSE_HTML = """
This product includes GeoLite2 data created by MaxMind, available from
<a href="https://www.maxmind.com">https://www.maxmind.com</a>.
"""

    _LICENSE_TXT = """
This product includes GeoLite2 data created by MaxMind, available from
https://www.maxmind.com.
"""

    _NO_API_KEY_MSSG = """
No API Key was found to download the Maxmind GeoIPLite database.
If you do not have an account, go here to create one and obtain and API key.
https://www.maxmind.com/en/geolite2/signup

Add this API key to your msticpyconfig.yaml
https://msticpy.readthedocs.io/en/latest/data_acquisition/GeoIPLookups.html#maxmind-geo-ip-lite-lookup-class.
Alternatively, you can pass this to the GeoLiteLookup class when creating it:
>>> iplookup = GeoLiteLookup(api_key="your_api_key")
"""
    _UNSET_PATH = "~~UNSET~~"

    def __init__(
        self,
        api_key: Optional[str] = None,
        db_folder: Optional[str] = None,
        force_update: bool = False,
        auto_update: bool = True,
        debug: bool = False,
    ):
        r"""
        Return new instance of GeoLiteLookup class.

        Parameters
        ----------
        api_key : str, optional
            Default is None - use configuration value from msticpyconfig.yaml.
            API Key from MaxMind -
            Read more about GeoLite2 : https://dev.maxmind.com/geoip/geoip2/geolite2/
            Sign up for a MaxMind account:
            https://www.maxmind.com/en/geolite2/signup
            Set your password and create a license key:
            https://www.maxmind.com/en/accounts/current/license-key
        db_folder: str, optional
            Provide absolute path to the folder containing MMDB file
            (e.g. '/usr/home' or 'C:/maxmind').
            If no path provided, it is set to download to .msticpy/GeoLite2
            under user`s home directory.
        force_update : bool, optional
            Force update can be set to true or false. depending on it,
            new download request will be initiated.
        auto_update: bool, optional
            Auto update can be set to true or false. depending on it,
            new download request will be initiated if age criteria is matched.
        debug : bool, optional
            Print additional debugging information, default is False.

        """
        self._debug = debug
        if self._debug:
            self._debug_init_state(api_key, db_folder, force_update, auto_update)
        self.settings: Optional[ProviderSettings] = None
        self._api_key: Optional[str] = api_key or None

        self._db_folder: str = db_folder or self._UNSET_PATH
        self._force_update = force_update
        self._auto_update = auto_update
        self._db_path: Optional[str] = None
        self._reader: Any = None

    def close(self):
        """Close an open GeoIP DB."""
        if self._reader:
            try:
                self._reader.close()
            except Exception as err:  # pylint: disable=broad-except
                print(f"Exception when trying to close GeoIP DB {err}")

    def lookup_ip(  # noqa: MC0001
        self,
        ip_address: str = None,
        ip_addr_list: Iterable = None,
        ip_entity: IpAddress = None,
    ) -> Tuple[List[Any], List[IpAddress]]:
        """
        Lookup IP location from GeoLite2 data created by MaxMind.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Any], List[IpAddress]]
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """
        self._check_initialized()
        ip_list = self._ip_params_to_list(ip_address, ip_addr_list, ip_entity)

        output_raw = []
        output_entities = []
        for ip_input in ip_list:
            geo_match = None
            try:
                ip_type = get_ip_type(ip_input)
            except ValueError:
                ip_type = "Invalid IP Address"
            if ip_type != "Public":
                geo_match = self._get_geomatch_non_public(ip_type)
            else:
                try:
                    geo_match = self._reader.city(ip_input).raw
                except (AddressNotFoundError, AttributeError, ValueError):
                    continue
            if geo_match:
                output_raw.append(geo_match)
                output_entities.append(
                    self._create_ip_entity(ip_input, geo_match, ip_entity)
                )

        return output_raw, output_entities

    @staticmethod
    def _get_geomatch_non_public(ip_type):
        """Return placeholder record for non-public IP Types."""
        return {
            "country": {
                "iso_code": None,
                "names": {"en": f"{ip_type} address"},
            },
            "city": {"names": {"en": "location unknown"}},
        }

    @staticmethod
    def _create_ip_entity(
        ip_address: str, geo_match: Mapping[str, Any], ip_entity: IpAddress = None
    ) -> IpAddress:
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_address
        geo_entity = GeoLocation()
        geo_entity.CountryCode = geo_match.get("country", {}).get("iso_code", None)
        geo_entity.CountryName = (
            geo_match.get("country", {}).get("names", {}).get("en", None)
        )
        subdivs = geo_match.get("subdivisions", [])
        if subdivs:
            geo_entity.State = subdivs[0].get("names", {}).get("en", None)
        geo_entity.City = geo_match.get("city", {}).get("names", {}).get("en", None)
        geo_entity.Longitude = geo_match.get("location", {}).get("longitude", None)
        geo_entity.Latitude = geo_match.get("location", {}).get("latitude", None)
        ip_entity.Location = geo_entity
        return ip_entity

    def _check_initialized(self):
        """Check if DB reader open with a valid database."""
        if self._reader and self.settings:
            return

        self.settings = _get_geoip_provider_settings("GeoIPLite")
        self._api_key = self._api_key or self.settings.args.get("AuthKey")

        self._db_folder: str = (
            self._db_folder
            if self._db_folder != self._UNSET_PATH
            else self.settings.args.get("DBFolder", self._DB_HOME)  # type: ignore
        )
        self._db_folder = str(Path(self._db_folder).expanduser())  # type: ignore
        self._check_and_update_db()
        self._db_path = self._get_geoip_db_path()
        if self._debug:
            self._debug_open_state()

        if not self._db_path:
            self._raise_no_db_error()
        self._reader = geoip2.database.Reader(self._db_path)

    def _check_and_update_db(self):
        """
        Check the age of geo ip database file and download if it older than 30 days.

        User can set auto_update or force_update to True or False to
        override auto-download behavior.

        """
        geoip_db_path = self._get_geoip_db_path()
        self._pr_debug(f"Checking geoip DB {geoip_db_path}")
        self._pr_debug(f"Download URL is {self._MAXMIND_DOWNLOAD}")
        if geoip_db_path is None:
            print(
                "No local Maxmind City Database found. ",
                f"Attempting to downloading new database to {self._db_folder}",
            )
            self._download_and_extract_archive()
        else:
            # Create a reader object to retrive db info and build date
            # to check age from build_epoch property.
            with geoip2.database.Reader(geoip_db_path) as reader:
                last_mod_time = datetime.fromtimestamp(
                    reader.metadata().build_epoch, tz=timezone.utc
                )

            # Check for out of date DB file according to db_age
            db_age = datetime.now(timezone.utc) - last_mod_time
            db_updated = True
            if db_age > timedelta(30) and self._auto_update:
                print(
                    "Latest local Maxmind City Database present is older than 30 days.",
                    f"Attempting to download new database to {self._db_folder}",
                )
                if not self._download_and_extract_archive():
                    self._geolite_warn("DB download failed")
                    db_updated = False
            elif self._force_update:
                print(
                    "force_update is set to True.",
                    f"Attempting to download new database to {self._db_folder}",
                )
                if not self._download_and_extract_archive():
                    self._geolite_warn("DB download failed")
                    db_updated = False
            if not db_updated:
                self._geolite_warn(
                    "Continuing with cached database. Results may inaccurate."
                )

    # noqa: MC0001
    def _download_and_extract_archive(self) -> bool:  # noqa: MC0001
        """
        Download file from the given URL and extract if it is archive.

        Returns
        -------
        bool :
            True if download_success

        """
        if not self._api_key:
            return False
        url = self._MAXMIND_DOWNLOAD.format(license_key=self._api_key)

        if not Path(self._db_folder).exists():
            # using makedirs to create intermediate-level dirs to contain self._dbfolder
            Path(self._db_folder).mkdir(exist_ok=True, parents=True)
        # build a temp file name for the archive download
        rand_int = random.randint(10000, 99999)  # nosec
        db_archive_path = Path(self._db_folder).joinpath(
            self._DB_ARCHIVE.format(rand=rand_int)
        )
        self._pr_debug(f"Downloading GeoLite DB: {db_archive_path}")
        try:
            # wait a small rand amount of time in case multiple procs try
            # to download simultaneously
            sleep(rand_int / 500000)
            if list(Path(self._db_folder).glob(self._DB_ARCHIVE.format(rand="*"))):
                # Some other process is downloading, wait a little then return
                sleep(3)
                return True
            with httpx.stream(
                "GET", url, timeout=get_http_timeout(), headers=mp_ua_header()
            ) as response:
                print("Downloading and extracting GeoLite DB archive from MaxMind....")
                with open(db_archive_path, "wb") as file_hdl:
                    for chunk in response.iter_bytes(chunk_size=10000):
                        file_hdl.write(chunk)
                        file_hdl.flush()
            self._pr_debug(f"Downloaded GeoLite DB: {db_archive_path}")
        except httpx.HTTPError as http_err:
            self._geolite_warn(
                f"HTTP error occurred trying to download GeoLite DB: {http_err}"
            )
        # pylint: disable=broad-except
        except Exception as err:
            self._geolite_warn(
                f"Other error occurred trying to download GeoLite DB: {err}"
            )
        # pylint: enable=broad-except
        else:
            # no exceptions so extract the archive contents
            try:
                self._extract_to_folder(db_archive_path)
                print(
                    "Extraction complete. Local Maxmind city DB:", f"{db_archive_path}"
                )
                return True
            except PermissionError as err:
                self._geolite_warn(
                    f"Cannot overwrite GeoIP DB file: {db_archive_path}."
                    + " The file may be in use or you do not have"
                    + f" permission to overwrite.\n - {err}"
                )
            except Exception as err:  # pylint: disable=broad-except
                # There are several exception types that might come from
                # unpacking a tar.gz
                self._geolite_warn(
                    f"Error writing GeoIP DB file: {db_archive_path} - {err}"
                )
        finally:
            if db_archive_path.is_file():
                self._pr_debug(f"Removing temp file {db_archive_path}")
                db_archive_path.unlink()
        return False

    def _extract_to_folder(self, db_archive_path: Path):
        self._pr_debug(f"Extracting tarfile {db_archive_path}")
        temp_folder: Optional[Path] = None
        with tarfile.open(db_archive_path) as tar_archive:
            for member in tar_archive.getmembers():
                if not member.isreg():
                    # Skip the dirs to extract only file objects
                    continue
                self._pr_debug(f"extracting {member} to {self._db_folder}")
                tar_archive.extract(member, self._db_folder)
                # The files are extracted to a subfolder (with a date in the name)
                # We want to move these into the main folder above this.
                targetname = Path(member.name).name
                if targetname != member.name:
                    # if target name is not already in self._dbfolder
                    # move it to there
                    target_file_path = Path(self._db_folder).joinpath(member.name)
                    temp_folder = target_file_path.parent
                    target_file_path.replace(Path(self._db_folder).joinpath(targetname))
                    self._pr_debug(
                        f"Moving to {Path(self._db_folder).joinpath(targetname)}"
                    )
        # delete contents of temp folder and remove it
        if temp_folder:
            for temp_file in temp_folder.glob("*"):
                temp_file.unlink()
            self._pr_debug(f"Removing temp path {temp_folder}")
            temp_folder.rmdir()

    def _get_geoip_db_path(self) -> Optional[str]:
        """
        Get the correct path containing GeoLite City Database.

        Returns
        -------
        Optional[str]
            Returns the absolute path of local maxmind geolite city
            database after control flow logic.

        """
        latest_db_path = Path(self._db_folder or ".").joinpath(self._DB_FILE)
        return str(latest_db_path) if latest_db_path.is_file() else None

    def _pr_debug(self, *args):
        """Print out debug info."""
        if self._debug:
            print(*args)

    def _geolite_warn(self, mssg: str):
        self._pr_debug(mssg)
        warnings.warn(
            f"GeoIpLookup: {mssg}",
            UserWarning,
        )

    def _raise_no_db_error(self):
        raise MsticpyUserConfigError(
            "No usable GeoIP Database could be found.",
            (
                "Check that you have correctly configured the Maxmind API key in "
                "msticpyconfig.yaml."
            ),
            (
                "If you are using a custom DBFolder setting in your config, "
                + f"check that this is a valid path: {self._db_folder}."
            ),
            (
                "If you edit your msticpyconfig to change this setting run the "
                "following commands to reload your settings and retry:"
                "    import msticpy"
                "    msticpy.settings.refresh_config()"
            ),
            help_uri=(
                "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
                + "GeoIPLookups.html#maxmind-geo-ip-lite-lookup-class"
            ),
            service_uri="https://www.maxmind.com/en/geolite2/signup",
            title="Maxmind GeoIP database not found",
        )

    def _debug_open_state(self):
        dbg_api_key = (
            "None"
            if self._api_key is None
            else self._api_key[:4] + "*" * (len(self._api_key) - 4)
        )
        self._pr_debug(f"__init__ values (inc settings): api_key={dbg_api_key}")
        self._pr_debug(f"    db_folder={self._db_folder}")
        self._pr_debug(f"    force_update={self._force_update}")
        self._pr_debug(f"    auto_update={self._auto_update}")
        self._pr_debug(f"    dbpath={self._db_path}")
        self._pr_debug(f"Using config file: {current_config_path()}")

    def _debug_init_state(self, api_key, db_folder, force_update, auto_update):
        dbg_api_key = (
            "None" if api_key is None else api_key[:4] + "*" * (len(api_key) - 4)
        )
        self._pr_debug(f"__init__ params: api_key={dbg_api_key}")
        self._pr_debug(f"    db_folder={db_folder}")
        self._pr_debug(f"    force_update={force_update}")
        self._pr_debug(f"    auto_update={auto_update}")


def _get_geoip_provider_settings(provider_name: str) -> ProviderSettings:
    """
    Return settings for a provider.

    Parameters
    ----------
    provider_name : str
        Name of the provider.

    Returns
    -------
    ProviderSettings
        Settings for the provider.

    """
    settings = get_provider_settings(config_section="OtherProviders")
    if provider_name in settings:
        return settings[provider_name]
    return ProviderSettings(  # type: ignore[call-arg]
        name=provider_name, description="Not found."
    )


@export
def entity_distance(ip_src: IpAddress, ip_dest: IpAddress) -> float:
    """
    Return distance between two IP Entities.

    Parameters
    ----------
    ip_src : IpAddress
        Source/Origin IpAddress Entity
    ip_dest : IpAddress
        Destination IpAddress Entity

    Returns
    -------
    float
        Distance in kilometers.

    Raises
    ------
        AttributeError
            If either entity has no location information

    """
    if not ip_src.Location or not ip_dest.Location:
        raise AttributeError(
            "Source and destination entities must have defined Location properties."
        )

    return geo_distance(
        origin=(ip_src.Location.Latitude, ip_src.Location.Longitude),
        destination=(ip_dest.Location.Latitude, ip_dest.Location.Longitude),
    )


_EARTH_RADIUS_KM = 6371  # km


@export
def geo_distance(
    origin: Tuple[float, float], destination: Tuple[float, float]
) -> float:
    """
    Calculate the Haversine distance.

    Parameters
    ----------
    origin : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.
    destination : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.

    Returns
    -------
    float
        Distance in kilometers.

    Examples
    --------
    >>> origin = (48.1372, 11.5756)  # Munich
    >>> destination = (52.5186, 13.4083)  # Berlin
    >>> round(geo_distance(origin, destination), 1)
    504.2

    Notes
    -----
        Author: Martin Thoma - stackoverflow

    """
    orig_lat, orig_lon = origin
    dest_lat, dest_lon = destination

    ang_dist_lat = math.radians(dest_lat - orig_lat)
    ang_dist_lon = math.radians(dest_lon - orig_lon)
    hav_a = (math.sin(ang_dist_lat / 2) * math.sin(ang_dist_lat / 2)) + (
        math.cos(math.radians(orig_lat))
        * math.cos(math.radians(dest_lat))
        * math.sin(ang_dist_lon / 2)
        * math.sin(ang_dist_lon / 2)
    )
    hav_c = 2 * math.atan2(math.sqrt(hav_a), math.sqrt(1 - hav_a))
    return _EARTH_RADIUS_KM * hav_c



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Cybereason Driver class."""
import datetime as dt
import json
import logging
import re
from asyncio import Future, as_completed
from concurrent.futures import ThreadPoolExecutor
from functools import partial, singledispatch
from typing import Any, Dict, List, Optional, Tuple, Union

import httpx
import pandas as pd
from tqdm.auto import tqdm

from ..._version import VERSION
from ...common.exceptions import MsticpyUserConfigError
from ...common.provider_settings import ProviderArgs, get_provider_settings
from ...common.utility import mp_ua_header
from ..core.query_defns import Formatters
from ..core.query_provider_connections_mixin import _get_event_loop
from .driver_base import DriverBase, DriverProps, QuerySource

__version__ = VERSION
__author__ = "Florian Bracq"

logger = logging.getLogger(__name__)

_HELP_URI = (
    "https://msticpy.readthedocs.io/en/latest/data_acquisition/DataProviders.html"
)


CybereasonSettings = Dict[str, Dict[str, Union[str, ProviderArgs]]]


# pylint: disable=too-many-instance-attributes
class CybereasonDriver(DriverBase):
    """Class to interact with Cybereason."""

    CONFIG_NAME = "Cybereason"

    _CONFIG_NAME_MAP = {
        "tenant_id": ("tenantid", "tenant_id"),
        "client_id": ("clientid", "client_id"),
        "client_secret": ("clientsecret", "client_secret"),
    }

    def __init__(self, **kwargs):
        """
        Instantiate Cybereason driver.

        Additional Parameters
        ---------------------
        timeout : int
            Query timeout in seconds. Defaults to 2min
        max_results : int
            Number of total results to return. Defaults to 1000
            Max is 10,000.
        page_size : int
            Number of results to return per page. Defaults to 100

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        super().__init__(**kwargs)
        timeout = kwargs.get("timeout", 120)  # 2 minutes in milliseconds
        logger.debug("Set timeout to %d", timeout)
        max_results = min(kwargs.get("max_results", 100000), 100000)
        logger.debug("Set maximum results to %d", max_results)
        self.base_url: str = "https://{tenant_id}.cybereason.net"
        self.auth_endpoint: str = "/login.html"
        self.req_body: Dict[str, Any] = {
            "queryPath": [],
            "totalResultLimit": max_results,
            "perGroupLimit": 100,
            "perFeatureLimit": 100,
            "templateContext": "SPECIFIC",
            "queryTimeout": timeout * 1000,
            "customFields": [],
        }
        self.search_endpoint: str = "/rest/visualsearch/query/simple"
        self._loaded = True
        self.client = httpx.Client(
            follow_redirects=True,
            timeout=self.get_http_timeout(timeout=timeout, def_timeout=120),
            headers=mp_ua_header(),
        )
        self.set_driver_property(
            DriverProps.FORMATTERS,
            {
                Formatters.PARAM_HANDLER: self._custom_param_handler,
                Formatters.DATETIME: self._format_datetime,
                Formatters.LIST: self._format_list,
            },
        )

        self.set_driver_property(DriverProps.SUPPORTS_THREADING, value=True)
        self.set_driver_property(
            DriverProps.MAX_PARALLEL, value=kwargs.get("max_threads", 4)
        )
        self._debug = kwargs.get("debug", False)

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err(self.__class__.__name__)

        page_size = min(kwargs.get("page_size", 2000), 4000)
        logger.debug("Set page size to %d", page_size)
        json_query = json.loads(query)
        body = {**self.req_body, **json_query}

        # The query must be executed at least once to retrieve the number
        # of results and the pagination token.
        response = self.__execute_query(body, page_size=page_size)

        total_results = response["data"]["totalResults"]
        pagination_token = response["data"]["paginationToken"]
        results: Dict[str, Any] = response["data"]["resultIdToElementDataMap"]

        logger.debug("Retrieved %d/%d results", len(results), total_results)

        df_result: pd.DataFrame = None  # type: ignore

        if len(results) < total_results:
            df_result = self._exec_paginated_queries(
                body=body,
                page_size=page_size,
                pagination_token=pagination_token,
                total_results=total_results,
            )
        else:
            df_result = self._format_result_to_dataframe(result=response)

        return df_result

    def _exec_paginated_queries(
        self,
        body: Dict[str, Any],
        page_size: int,
        pagination_token: str,
        total_results: int,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Return results of paginated queries.

        Parameters
        ----------
        body : Dict[str, Any]
            The body of the query to execute.

        Additional Parameters
        ----------
        progress: bool, optional
            Show progress bar, by default True
        retry_on_error: bool, optional
            Retry failed queries, by default False
        **kwargs : Dict[str, Any]
            Additional keyword arguments to pass to the query method.

        Returns
        -------
        pd.DataFrame
            The concatenated results of all the paginated queries.

        Notes
        -----
        This method executes the specified query multiple times to retrieve
        all the data from paginated results.
        The queries are executed asynchronously.

        """
        progress = kwargs.pop("progress", True)
        retry = kwargs.pop("retry_on_error", False)

        query_tasks = self._create_paginated_query_tasks(
            body=body,
            page_size=page_size,
            pagination_token=pagination_token,
            total_results=total_results,
        )

        logger.info("Running %s paginated queries.", len(query_tasks))
        event_loop = _get_event_loop()
        return event_loop.run_until_complete(
            self.__run_threaded_queries(query_tasks, progress, retry)
        )

    def connect(
        self,
        connection_str: Optional[str] = None,
        **kwargs,
    ):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str: Optional[str], optional
            Connect to a data source
        instance : Optional[str], optional
            Optional name of configuration instance - this
            is added as a prefix to the driver configuration key name
            when searching for configuration in the msticpyconfig.yaml

        Notes
        -----
        Connection string fields:
            instance
            client_id
            client_secret

        """
        cs_dict: Dict[str, Any] = {}

        self._instance = kwargs.pop("instance", None)
        cs_dict = CybereasonDriver._get_driver_settings(
            self.CONFIG_NAME, self._instance
        )
        # let user override config settings with function kwargs
        cs_dict.update(kwargs)

        missing_settings = [
            setting
            for setting in ("tenant_id", "client_id", "client_secret")
            if setting not in cs_dict
        ]
        if missing_settings:
            raise MsticpyUserConfigError(
                "You must supply the following required connection parameter(s)",
                "to the connect function or add them to your msticpyconfig.yaml.",
                ", ".join(f"'{param}'" for param in missing_settings),
                title="Missing connection parameters.",
                help_uri=("Connecting to OData sources.", _HELP_URI),
            )

        # self.auth_endpoint and self.req_body are correctly set in concrete
        # instances __init__
        self.client.base_url = httpx.URL(
            self.base_url.format(tenant_id=cs_dict["tenant_id"])
        )
        req_body: Dict[str, str] = {
            "username": cs_dict["client_id"],
            "password": cs_dict["client_secret"],
        }

        # Authenticate and obtain cookie for future calls
        response = self.client.post(self.auth_endpoint, data=req_body)
        response.raise_for_status()

        print("Connected.")
        self._connected = True
        self.current_connection = self._instance

        return self._connected

    @staticmethod
    def _flatten_result(entry: Dict[str, Any]) -> Dict[str, Any]:
        """
        Flatten Cybereason result to a format that can be handled by pandas.

        Parameters
        ----------
        entry: Dict[str, Any]
            Entry to flatten

        Returns
        -------
        Dict[str, Any]

        """
        result = {}
        # Retrieve simpleValues and add them to the output
        simple_values: Dict[str, Any] = entry.get("simpleValues", {})
        result = CybereasonDriver._flatten_simple_values(simple_values)

        elt_value = entry.get("elementValues", {})  # List or Dict
        result.update(**CybereasonDriver._flatten_element_values(elt_value))
        return result

    @staticmethod
    def _flatten_simple_values(simple_values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Flatten "simpleValues from Cybereason result to a format that can be handled by pandas.

        Parameters
        ----------
        simple_values: Dict[str, Any]
            Entry to flatten

        Returns
        -------
        Dict[str, Any]

        """
        result: Dict[str, Any] = {}
        for name, values in simple_values.items():
            if not values["values"]:
                return result
            result[name] = list(
                {
                    (
                        CybereasonDriver._format_to_datetime(int(value))
                        if "Time" in name
                        else value.strip().rstrip("\x00")
                    )
                    for value in values["values"]
                }
            )
            if values["totalValues"] == 1:
                result[name] = result[name][0]

        return result

    @staticmethod
    def _flatten_element_values(
        element_values: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> Dict[str, Any]:
        """
        Flatten "elementValues from Cybereason result to a format that can be handled by pandas.

        Parameters
        ----------
        element_values: Union[Dict[str, Any], List[str]]
            Entry to flatten

        Returns
        -------
        Dict[str, Any]

        """
        result = {}
        if isinstance(element_values, list):
            for values in element_values:
                result[values["elementType"]] = values["name"]
                result[f"{values['elementType']}.guid"] = values["guid"]
        elif isinstance(element_values, dict):
            for key, values in element_values.items():
                flattened = CybereasonDriver._flatten_result(values)
                if flattened:
                    for subkey, subvalues in flattened.items():
                        result[f"{key}.{subkey}"] = subvalues
        return result

    def _create_paginated_query_tasks(
        self,
        body: Dict[str, Any],
        page_size: int,
        pagination_token: str,
        total_results: int,
    ) -> Dict[str, partial]:
        """Return dictionary of partials to execute queries."""
        # Compute the number of queries to execute
        total_pages = total_results // page_size + 1
        # The first query (page 0) as to be re-run due to a bug in
        # Cybereason API. The first query returns less results than the page size
        # when executed without a pagination token.
        return {
            f"{page}": partial(
                self.__execute_query,
                body=body,
                page_size=page_size,
                pagination_token=pagination_token,
                page=page,
            )
            for page in range(0, total_pages)
        }

    def __execute_query(
        self,
        body: Dict[str, Any],
        page: int = 0,
        page_size: int = 2000,
        pagination_token: str = None,
    ) -> Dict[str, Any]:
        """
        Run query with pagination enabled.

        Parameters
        ----------
        body: Dict[str, Any]
            Body of the HTTP Request
        page_size: int
            Size of the page for results
        page: int
            Page number to query
        pagination_token: str
            Token of the current search

        Returns
        -------
        Dict[str, Any]

        """
        if pagination_token:
            pagination = {
                "pagination": {
                    "pageSize": page_size,
                    "page": page + 1,
                    "paginationToken": pagination_token,
                    "skip": page * page_size,
                }
            }
            headers = {"Pagination-Token": pagination_token}
        else:
            pagination = {"pagination": {"pageSize": page_size}}
            headers = {}
        params = {"page": page, "itemsPerPage": page_size}
        status = None
        while status != "SUCCESS":
            response = self.client.post(
                self.search_endpoint,
                json={**body, **pagination},
                headers=headers,
                params=params,
            )
            response.raise_for_status()
            json_result = response.json()
            status = json_result["status"]
        return json_result

    async def __run_threaded_queries(
        self,
        query_tasks: Dict[str, partial],
        progress: bool = True,
        retry: bool = False,
    ) -> pd.DataFrame:
        logger.info("Running %d threaded queries.", len(query_tasks))
        event_loop = _get_event_loop()
        with ThreadPoolExecutor(max_workers=4) as executor:
            results: List[pd.DataFrame] = []
            failed_tasks: Dict[str, Future] = {}
            thread_tasks = {
                query_id: event_loop.run_in_executor(executor, query_func)
                for query_id, query_func in query_tasks.items()
            }
            if progress:
                task_iter = tqdm(
                    as_completed(thread_tasks.values()),
                    unit="paginated-queries",
                    desc="Running",
                )
            else:
                task_iter = as_completed(thread_tasks.values())
            ids_and_tasks = dict(zip(thread_tasks, task_iter))
            for query_id, thread_task in ids_and_tasks.items():
                try:
                    result = await thread_task
                    df_result = self._format_result_to_dataframe(result)
                    logger.info("Query task '%s' completed successfully.", query_id)
                    results.append(df_result)
                except Exception:  # pylint: disable=broad-except
                    logger.warning(
                        "Query task '%s' failed with exception", query_id, exc_info=True
                    )
                    failed_tasks[query_id] = thread_task

            if retry and failed_tasks:
                for query_id, thread_task in failed_tasks.items():
                    try:
                        logger.info("Retrying query task '%s'", query_id)
                        result = await thread_task
                        df_result = self._format_result_to_dataframe(result)
                        results.append(df_result)
                    except Exception:  # pylint: disable=broad-except
                        logger.warning(
                            "Retried query task '%s' failed with exception",
                            query_id,
                            exc_info=True,
                        )
            # Sort the results by the order of the tasks
            results = [result for _, result in sorted(zip(thread_tasks, results))]
        return pd.concat(results, ignore_index=True)

    # pylint: disable=too-many-branches
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) and
            Kql ResultSet.

        """
        raise NotImplementedError(f"Not supported for {self.__class__.__name__}")

    # Parameter Formatting method
    @staticmethod
    def _format_datetime(date_time: dt.datetime) -> int:
        """Return datetime formatted as timestamp in milliseconds."""
        return int(date_time.timestamp() * 1000)

    @staticmethod
    def _format_list(value: List) -> List:
        """Return list as itself."""
        return value

    # Parameter Formatting method
    @staticmethod
    def _format_to_datetime(timestamp: int) -> Union[dt.datetime, int]:
        """Return datetime from a timestamp in milliseconds."""
        try:
            return dt.datetime.fromtimestamp(timestamp // 1000)
        except TypeError:
            return timestamp

    @staticmethod
    def _format_result_to_dataframe(result: Dict[str, Any]) -> pd.DataFrame:
        """Return a dataframe from a cybereason result object."""
        df_result = [
            dict(
                CybereasonDriver._flatten_result(values),
                **{"resultId": result_id},
            )
            for result_id, values in result["data"]["resultIdToElementDataMap"].items()
        ]
        return pd.json_normalize(df_result)

    # Retrieve configuration parameters with aliases
    @staticmethod
    def _map_config_dict_name(config_dict: Dict[str, str]):
        """Map configuration parameter names to expected values."""
        mapped_dict = config_dict.copy()
        for provided_name in config_dict:
            for req_name, alternates in CybereasonDriver._CONFIG_NAME_MAP.items():
                if provided_name.casefold() in alternates:
                    mapped_dict[req_name] = config_dict[provided_name]
                    break
        return mapped_dict

    # Read values from configuration
    @staticmethod
    def _get_driver_settings(
        config_name: str, instance: Optional[str] = None
    ) -> Dict[str, str]:
        """Try to retrieve config settings for Cybereason drivers."""
        config_key = f"{config_name}-{instance}" if instance else config_name
        drv_config = get_provider_settings("DataProviders").get(config_key)
        app_config: Dict[str, str] = {}
        if drv_config:
            app_config = dict(drv_config.args)

        if not app_config:
            return {}
        # map names to allow for different spellings
        return CybereasonDriver._map_config_dict_name(app_config)

    @staticmethod
    def _custom_param_handler(query: str, param_dict: Dict[str, Any]) -> str:
        """Replace parameters in query template for Cybereason JSON queries."""
        query_dict = json.loads(query)

        return json.dumps(_recursive_find_and_replace(query_dict, param_dict))


@singledispatch
def _recursive_find_and_replace(
    parameters: Union[str, Dict, List], param_dict: Dict[str, Any]
):
    """Recursively find and replace parameters from query."""
    if isinstance(parameters, (list, str, dict)):
        return _recursive_find_and_replace(parameters, param_dict)
    return parameters


@_recursive_find_and_replace.register(dict)
def _(parameters: Dict[str, Any], param_dict: Dict[str, Any]):
    return {
        parameter: _recursive_find_and_replace(value, param_dict)
        for parameter, value in parameters.items()
    }


@_recursive_find_and_replace.register(list)
def _(parameters: List, param_dict: Dict[str, Any]):
    result = [
        _recursive_find_and_replace(parameter, param_dict) for parameter in parameters
    ]
    if all(isinstance(values, list) for values in result):
        try:
            return sorted({value for values in result for value in values})
        except TypeError:
            # If we have a list with different types,convert all to string.
            return sorted({str(value) for values in result for value in values})
    return result


@_recursive_find_and_replace.register(str)
def _(parameters: str, param_dict: Dict[str, Any]):
    """Recursively find and replace parameters from query."""
    param_regex = r"{([^}]+)}"
    matches = re.match(param_regex, parameters)
    if matches:
        result = [param_dict.get(match, parameters) for match in matches.groups()]
        if len(result) == 1:
            return result[0]
        return result
    return parameters



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Security Graph OData Driver class."""
from typing import Any, Optional, Union

import pandas as pd

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...common.utility import export
from .odata_driver import OData, QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class SecurityGraphDriver(OData):
    """Driver to query security graph."""

    CONFIG_NAME = "MicrosoftGraph"
    _ALT_CONFIG_NAMES = ["SecurityGraphApp"]

    def __init__(self, connection_str: Optional[str] = None, **kwargs):
        """
        Instantiate MSGraph driver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        """
        super().__init__(**kwargs)
        az_cloud_config = AzureCloudConfig(cloud=kwargs.pop("cloud", None))
        self.scopes = ["User.Read"]
        self.api_root = az_cloud_config.endpoints.get("microsoftGraphResourceId")
        self.req_body = {
            "client_id": None,
            "client_secret": None,
            "grant_type": "client_credentials",
            "scope": f"{self.api_root}.default",
        }
        login_endpoint = az_cloud_config.authority_uri
        self.oauth_url = f"{login_endpoint}{{tenantId}}/oauth2/v2.0/token"
        self.api_ver = kwargs.get("api_ver", "v1.0")

        if connection_str:
            self.current_connection = connection_str
            self.connect(connection_str)

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        del query_source, kwargs
        return self.query_with_results(query, body=False)[0]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Visualization sub-package.

This contains the following functionality:

- entity_graph_tools - graphing and timeline support for entity visualization
- data_viewer - Bokeh-based data browser
- data_viewer_pane - PyVis Panel-based data browser
- foliummap - mapping using the folium package
- matrix_plot - matrix/interaction plots
- morph_charts - experimental morph charts support
- mp_pandas_plot - pandas `mp_plot` accessor for plotting functions
- process_tree - process tree visualization
- timeline - event timeline and timeline_values visualizations
- timeline_duration - timeline durations
- timeline_pd_accessor - (deprecated) timeline pd accessor
- timeseries - timeseries analysis visualization

"""
# flake8: noqa: F403
# pylint: disable=unused-import
from . import mp_pandas_plot

__all__ = ["mp_pandas_plot"]
